"""Module with possible events and classes/functions related to that."""

from typing import List, Dict, Callable, Tuple, Awaitable, Optional
from random import random
from abc import ABC
import logging

from .utils import read_only_properties
from .objects import Action, Attachment


Callback = Callable[["Event"], Awaitable]


_SLOTS: Tuple[str, ...] = ()


# ----------------------------------------------------------------------------
# Basic events


@read_only_properties("name", "group_id", "source")
class Event(ABC):
    """Base class for possible events."""

    __slots__ = (
        "_app", "_callbacks", "_callbacks_index",
        "name", "group_id", "source",
    )

    def __init__(self, name: str, group_id: int, source: Dict):
        self._app = None
        self._callbacks: List[Callback] = []
        self._callbacks_index: int = -1

        #: Event's internal name (like "vk:<event's name>")
        self.name: str = name

        #: Event's group id.
        self.group_id = group_id
        #: Raw object of event.
        self.source = source

    def __str__(self):  # pragma: no cover
        return "<{}[{}] from {}>".format(
            self.__class__.__name__, self.name, self.group_id,
        )

    def initialize(self, app, callbacks: List[Callback]):
        """Set app amd callbacks that will process this event."""
        self._app = app
        self._callbacks = callbacks

    async def request(self, method, **kwargs):
        """Proxy for VkClient's `request`."""

        return await self._app.get_client(self.group_id).request(method, **kwargs)

    async def next(self):
        """Call next callback from the list."""

        self._callbacks_index += 1

        try:
            await self._callbacks[self._callbacks_index](self)
        except Exception:
            logging.exception("Callback for update")
            raise


class EventRaw(Event):
    """Raw event from vkontakte. Used for passing unknown events."""

    def __init__(self, group_id, source: Dict):
        super().__init__("vk:raw", group_id, source)


_SLOTS = (
    "id", "date", "peer_id", "from_id", "text", "random_id",
    "important", "payload", "out", "attachments", "action", "reply_message",
    "fwd_messages"
)

@read_only_properties(*_SLOTS)  #pylint: disable=too-many-instance-attributes
class MessageData(Event):
    """
    Class for storing data about personal message in vkontakte.
    Documentation: https://vk.com/dev/objects/message
    """

    __slots__ = _SLOTS

    def __init__(self, name, group_id, source: Dict):
        super().__init__(name, group_id, source)

        #: Message id
        self.id: int = source.get("id", 0)
        #: Message date in unixtime
        self.date: int = source.get("date", 0)
        #: Message peer id
        self.peer_id: int = source.get("peer_id", 0)
        #: Message from id
        self.from_id: int = source.get("from_id", 0)
        #: Message text if present
        self.text: str = source.get("text", "")
        #: Message random id
        self.random_id: int = source.get("random_id", 0)
        #: Message important flag
        self.important: bool = source.get("important", False)
        #: Message payload
        self.payload: str = source.get("payload", "")
        #: Message out flag
        self.out: bool = source.get("out", False)

        #: Message action if present
        self.action: Optional["Action"] = None
        if "action" in source:
            self.action = Action(source["action"])

        #: Message reply message if present
        self.reply_message: Optional[MessageData] = None
        if "reply_message" in source:
            self.reply_message = MessageData("", 0, source["reply_message"])

        #: Forwarded messages
        self.fwd_messages: Tuple[MessageData, ...] = tuple(
            MessageData("", 0, s) for s in source.get("fwd_messages", ())
        )

        #: Message attachments
        self.attachments: Tuple[Attachment, ...] = tuple(
            Attachment(s) for s in source.get("attachments", ())
        )

    async def response(self, message: str):
        """Response to channel event was received from."""

        return await self.request(
            "messages.send",
            message=message,
            peer_id=self.peer_id,
            random_id=int(random() * 4294967296) - 2147483648
        )


# ----------------------------------------------------------------------------
# [Vkontakte events](https://vk.com/dev/groups_events)


class MessageNew(MessageData):
    """Vkontakte "message_new" event."""

    def __init__(self, group_id, source: Dict):
        super().__init__("vk:message_new", group_id, source)



"""Library for interacting with Vkontakte."""

import logging

from .vkpore import Vkpore
from .vkclient import VkClient


logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [ %(levelname)s ] %(message)s'
)



"""Module with core class for organizing event flow."""

from typing import List, Dict, Awaitable, Optional, Iterable, Tuple
from random import choice
from asyncio import AbstractEventLoop as AEL
import asyncio
import logging

from aiohttp import ClientSession

from .vkclient import VkClient
from .events import Event, Callback, MessageNew, EventRaw
from .utils import wait_with_stopped


class Vkpore():
    """
    Class for receiving events, calling methods, callback registration
    and execution. You can specify loop and session to use.
    """

    def __init__(self, tokens: Iterable[str], loop: AEL = None,
                 session: ClientSession = None):
        self._loop: AEL = loop or asyncio.get_event_loop()
        self._callbacks: Dict[str, List[Callback]] = {}
        self._futures: List[asyncio.Future] = []

        self._tokens: Tuple[str, ...] = tuple(tokens)
        self._clients: Dict[int, List[VkClient]] = {}

        self._stopped: asyncio.Event = asyncio.Event()
        self._stopped.set()

        self._session: Optional[ClientSession] = session

    def get_client(self, group_id):
        """Return random client for specified group_id."""

        clients = self._clients.get(group_id)

        if not clients:
            return None

        return choice(clients)

    @staticmethod
    def _get_event_class(update_type):  # pragma: no cover
        if update_type == "message_new":
            return MessageNew

        return EventRaw

    async def _longpoll_loop(self, group_id):
        get_updates = self.get_client(group_id).longpoll()

        stopped = asyncio.Task(self._stopped.wait())

        while True:
            updates = await wait_with_stopped(get_updates(), stopped)

            if updates is None:
                break

            for update in updates:
                event_class = self._get_event_class(update["type"])

                self.dispatch(event_class(group_id, update["object"]))

    async def start(self):
        """Start application related loops and perform initializations."""

        if self._session is None:
            self._session = ClientSession()

        # Create and inititalize clients
        for token in self._tokens:
            client = VkClient(token, self._session, self._loop)

            await client.initialize()

            if client.group_id not in self._clients:
                self._clients[client.group_id] = []

            self._clients[client.group_id].append(client)

        # Application not is considered running
        self._stopped.clear()

        # Create and start loops for receiving updates for groups
        for group_id in self._clients:
            self._futures.append(
                asyncio.ensure_future(
                    self._longpoll_loop(group_id),
                    loop=self._loop
                )
            )

        # Start execute loops for clients
        for clients in self._clients.values():
            for client in clients:
                client.start()

        logging.info("Started")

    async def stop(self):
        """
        Stop client loops, wait for all tasks to complete and close
        sessions.
        """

        logging.info("Stopping")

        self._stopped.set()

        # Wait for running callbacks to stop
        await asyncio.gather(*self._futures)

        # Wait for running loops to stop
        tasks = []

        for group in self._clients.values():
            for client in group:
                tasks.append(client.stop())

        await asyncio.gather(*tasks)

        # Close session
        if self._session:
            await self._session.close()

        logging.info("Stopped")

    def on(self, event: str):
        """Return decorator for subscribing callback to event."""

        def decorator(function: Callback):
            if event not in self._callbacks:
                self._callbacks[event] = []

            self._callbacks[event].append(function)

        return decorator

    def dispatch(self, event: Event):
        """Call callbacks for event."""

        callbacks = self._callbacks.get(event.name)

        if not callbacks or self._stopped.is_set():
            return None

        event.initialize(self, callbacks)

        future = asyncio.ensure_future(event.next(), loop=self._loop)

        self._futures.append(future)

        logging.debug("Dispatched event: %s", event)

        return future

    def run_until_complete(self, awaitable: Awaitable):  # pragma: no cover
        """Run specified awaitable in application's loop."""
        self._loop.run_until_complete(awaitable)

    def run(self):  # pragma: no cover
        """Run application and stop on KeyboardInterrupt."""
        try:
            self._loop.run_until_complete(self.start())
            self._loop.run_forever()
        except KeyboardInterrupt:
            self.run_until_complete(self.stop())



"""Useful helpers"""

from typing import Awaitable
import asyncio


async def wait_with_stopped(awaitable: Awaitable, stopped: Awaitable, loop=None):
    """
    Wait for awaitable or stopped to complete. If stopped was
    completed - awaitable is cancelled and None is returned. Otherwise
    return awaitable's result.
    """

    done, pending = await asyncio.wait(
        {awaitable, stopped}, return_when=asyncio.FIRST_COMPLETED, loop=loop,
    )

    if stopped in done:
        if pending:
            pending.pop().cancel()
        return None

    return done.pop().result()


def read_only_properties(*attrs):
    """Make passed attributes read-only"""
    def decorator(cls):
        original_setattr = cls.__setattr__

        def modified_setattr(self, name, value):
            if name in attrs and getattr(self, name, None) is not None:
                raise AttributeError("Can't modify '{}'".format(name))
            original_setattr(self, name, value)

        cls.__setattr__ = modified_setattr

        return cls
    return decorator



"""Module with classes related to interacting with Vkontakte."""

from typing import List, Dict, Union, Awaitable, Optional, Callable
from asyncio import Future, AbstractEventLoop as AEL
import asyncio
import logging
import json

from aiohttp import ClientSession, ClientError

from .utils import wait_with_stopped


class Request(Future):
    """Request in queue for execution."""

    __slots__ = ("method", "arguments")

    def __init__(self, method, arguments):
        super().__init__()

        self.method: str = method
        self.arguments: Dict[str, Union[str, int]] = arguments


class VkClient:  #pylint: disable=too-many-instance-attributes
    """Class for interacting with Vkontakte."""

    def __init__(self, token: str, session: ClientSession = None, loop: AEL = None):
        self._token: str = token
        self._loop: AEL = loop or asyncio.get_event_loop()
        self._session: ClientSession = session or ClientSession()
        self._api_url: str = "https://api.vk.com/method/{method}"
        self._version: str = "5.92"
        self._loop_pause: float = 1 / 19

        self._queue: asyncio.Queue = asyncio.Queue()
        self._running_loop: Optional[Awaitable] = None

        self._stopped: asyncio.Event = asyncio.Event()
        self._stopped.set()

        self._group_id: int = 0
        self._group_name: str = ""

    async def _execute_loop(self):
        stopped = asyncio.Task(self._stopped.wait())

        while True:
            await asyncio.sleep(self._loop_pause)

            request = await wait_with_stopped(self._queue.get(), stopped)

            if request is None:
                break

            requests: List[Request] = [request]

            for _ in range(24):
                try:
                    requests.append(self._queue.get_nowait())
                except asyncio.QueueEmpty:
                    break

            code = ["return ["]

            for request in requests:
                code.append("API.{}({}),".format(
                    request.method,
                    json.dumps(request.arguments, ensure_ascii=False),
                ))

            code.append("];")

            responses = await self.raw_request("execute", code="".join(code))

            if responses is None:
                responses = [False] * len(requests)

            for response, request in zip(responses, requests):
                try:
                    if response is False:
                        request.set_result(None)
                    else:
                        request.set_result(response)
                except asyncio.InvalidStateError:  # pragma: no cover
                    pass

    @property
    def group_id(self):
        """Client's group id"""
        return self._group_id

    @property
    def group_name(self):
        """Client's group name"""
        return self._group_name

    async def initialize(self, enable_longpoll=True):
        """Check token and get information about group."""

        resp = await self.raw_request("groups.getById")

        if not resp:
            raise ValueError("Initialization failed")

        self._group_id = resp[0]["id"]
        self._group_name = resp[0]["name"]

        logging.info(
            'Initialized "%s" [ https://vk.com/club%s ]',
            self.group_name, self.group_id,
        )

        if enable_longpoll:
            await self.raw_request(
                "groups.setLongPollSettings",
                group_id=self._group_id,
                api_version=self._version,
                enabled=1,
            )

    def start(self):
        """
        Start loop for executing requests with `request` method. Loop
        allows usage of batching and `execute` method. Returns awaitable
        with loop. Only one loop cann run for instance of client.
        """

        if self._running_loop:
            raise RuntimeError("Loop already running!")

        self._stopped.clear()

        self._running_loop = asyncio.ensure_future(
            self._execute_loop(), loop=self._loop
        )

        return self._running_loop

    async def stop(self):
        """
        Attempts to stop background loop and waits for it to actually stop.
        """

        self._stopped.set()

        if self._running_loop:
            await self._running_loop

    async def request(self, method: str, **kwargs):
        """
        Perform a request to method with arguments. Access token and version
        added explicitly, but you can override it with your arguments. Request
        if performed from background loop and is batched in order to use
        `execute` method. Returns response or None if error occured.
        """

        if not self._running_loop:
            raise RuntimeError("Loop for requests is not running!")

        request = Request(method, kwargs)
        await self._queue.put(request)
        return await request

    async def raw_request(self, method: str, **kwargs):
        """
        Perform a request to method with arguments. Access token and version
        added explicitly, but you can override it with your arguments.
        Returns response or None if error occured.
        """

        logging.debug("Request: [%s %s]", method, str(kwargs))

        arguments = {
            "v": self._version,
            "access_token": self._token,
            **{k: v for k, v in kwargs.items() if v is not None}
        }

        url = self._api_url.format(method=method)

        try:
            async with self._session.post(url, data=arguments) as raw_response:
                response = await raw_response.json(content_type=None)

                if not response or response.get("response", None) is None:
                    return None

                return response["response"]
        except (json.JSONDecodeError, ClientError):
            logging.exception("Performing raw request")
            return None

    def longpoll(self, default_longpoll=None) -> Callable[[], Callable[[], List[Dict]]]:
        """
        Return coroutine for receiving updates for current group from
        vkontakte using Bots Longpoll API.
        """

        if default_longpoll is None:
            longpoll: Dict[str, Union[int, str]] = {}
        else:
            longpoll = {**default_longpoll}

        async def refresh():
            """Get new values for longpolling."""

            response = await self.request(
                "groups.getLongPollServer", group_id=self.group_id
            )

            if response is not None:
                longpoll["server"] = response["server"]
                longpoll["key"] = response["key"]
                longpoll["ts"] = response["ts"]

        async def get():
            """Request new updates from Vkontakte."""

            while not longpoll:
                await refresh()

            arguments = {
                "ts": longpoll["ts"],
                "act": "a_check",
                "key": longpoll["key"],
                "wait": 25,
            }

            try:
                post = self._session.post(longpoll["server"], data=arguments)

                async with post as raw_response:
                    response = await raw_response.json(content_type=None)

            except (json.JSONDecodeError, ClientError):
                logging.exception("Longpoll request")
                longpoll.clear()
                return

            if "ts" in response:
                longpoll["ts"] = response["ts"]

            if "failed" in response:
                if response["failed"] != 1:
                    longpoll.clear()
                return

            return response["updates"]

        return get

    async def close_session(self):  # pragma: no cover
        """Close this client's session."""
        if self._session:
            await self._session.close()



"""Module with vkontakte API objects."""
from typing import Dict, Union, Optional, Tuple
from abc import ABC

from .utils import read_only_properties


SLOTS: Tuple[str, ...] = ()


@read_only_properties("type", "content")  # pylint: disable=too-few-public-methods
class Attachment:
    """Class for storing information about attachment."""

    __slots__ = ("type", "content", "content_raw",)

    def __init__(self, source: Dict):
        #: Attachment type.
        self.type: str = source.get("type", "")
        #: Raw object.
        self.content_raw: Dict = source.get(self.type, {})
        #: Supported attachment content class or None.
        self.content: Union[
            None, Photo, Video, Audio, Doc, Link, Sticker, Gift, Wall
        ] = None

        if self.type == "photo":
            self.content = Photo(self.content_raw)
        elif self.type == "video":
            self.content = Video(self.content_raw)
        elif self.type == "audio":
            self.content = Audio(self.content_raw)
        elif self.type == "doc":
            self.content = Doc(self.content_raw)
        elif self.type == "link":
            self.content = Link(self.content_raw)
        elif self.type == "sticker":
            self.content = Sticker(self.content_raw)
        elif self.type == "gift":
            self.content = Gift(self.content_raw)
        elif self.type == "wall":
            self.content = Wall(self.content_raw)


SLOTS = (
    "type", "source", "id", "owner_id", "access_key",
)


@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods
class AttachmentContent(ABC):
    """Class for storing information about attachment's content"""

    __slots__ = tuple(SLOTS)

    def __init__(self, _type: str, source: Dict):
        #: Content's type
        self.type: str = _type

        #: ID.
        self.id: int = int(source.get("id", 0))
        #: Owner id.
        self.owner_id: int = int(source.get("owner_id", 0))
        #: Access key if present.
        self.access_key: str = source.get("access_key", "")

        #: Raw source of content
        self.source: Dict = source

    @property
    def prepared(self) -> str:
        """
        Return string in format "<type><owner_id>_<media_id>".
        If attachment has "access_key" fields - it will be properly appended.
        """
        template = "{}{}_{}"

        if self.access_key:
            template = "{}{}_{}_" + self.access_key

        return template.format(self.type, self.owner_id, self.id)


SLOTS = (
    "album_id", "user_id", "text", "date", "sizes",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Photo(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/photo
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("photo", source)

        #: Album id.
        self.album_id: int = int(source.get("album_id", 0))
        #: Uploader's id (100 if uploaded by group)
        self.user_id: int = int(source.get("user_id", 0))
        #: Description
        self.text: str = source.get("text", "")
        #: Date of adding in Unixtime
        self.date: int = source.get("date", "")
        #: Tuple with photo's sizes
        self.sizes: Tuple[PhotoSize, ...] = tuple(
            PhotoSize(s) for s in source.get("sizes", ())
        )

    @property
    def uploaded_by_group(self) -> bool:
        """Return True if photo was uploaded by user."""
        return self.user_id == 100


@read_only_properties("type", "url", "width", "height")  # pylint: disable=too-few-public-methods
class PhotoSize:
    """
    Class for string information about photo size. Documentation:
    https://vk.com/dev/photo_sizes
    """

    __slots__ = ("type", "url", "width", "height")

    def __init__(self, source: Dict):
        #: Size type
        self.type = source.get("type", "")
        #: Image url
        self.url = source.get("url", "")
        #: Image width
        self.width = int(source.get("width", 0))
        #: Image height
        self.height = int(source.get("height", 0))


SLOTS = (
    "title", "description", "duration", "date",
    "adding_date", "views", "comments", "player",
    "first_frame", "photo",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Video(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    https://vk.com/dev/objects/video
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("video", source)

        #: Title
        self.title: str = source.get("title", "")
        #: Description
        self.description: str = source.get("description", "")
        #: Duration in seconds
        self.duration: int = int(source.get("duration", 0))
        #: Upload date in unixtime
        self.date: int = int(source.get("date", 0))
        #: Adding date in unixtime
        self.adding_date: int = int(source.get("adding_date", 0))
        #: Views count
        self.views: int = int(source.get("views", 0))
        #: Comments count
        self.comments: int = int(source.get("comments", 0))
        #: Player url
        self.player: str = source.get("player", "")
        #: Url to larges first frame
        self.first_frame: str = (
            source.get("first_frame_1280", "")
            or source.get("first_frame_800", "")
            or source.get("first_frame_640", "")
            or source.get("first_frame_320", "")
            or source.get("first_frame_130", "")
        )
        #: Url to largest cover
        self.photo: str = (
            source.get("photo_1280", "")
            or source.get("photo_800", "")
            or source.get("photo_640", "")
            or source.get("photo_320", "")
            or source.get("photo_130", "")
        )


SLOTS = (
    "artist", "title", "duration", "url", "lyrics_id", "album_id", "genre_id",
    "date", "is_hq", "is_explicit",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Audio(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    https://vk.com/dev/objects/audio
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("audio", source)

        #: Title
        self.title: str = source.get("title", "")
        #: Artist
        self.artist: str = source.get("artist", "")
        #: Duration in seconds
        self.duration: int = int(source.get("duration", 0))
        #: Upload date in unixtime
        self.date: int = int(source.get("date", 0))
        #: Album id
        self.album_id: int = int(source.get("album_id", 0))
        #: Url to .mo3 file
        self.url: str = source.get("url", "")
        #: Lyrics ID if present
        self.lyrics_id: int = int(source.get("lyrics_id", 0))
        #: Genre ID
        self.genre_id: int = int(source.get("genre_id", 0))
        #: High quality flag
        self.is_hq: bool = source.get("is_hq", False)
        #: Explicit flag
        self.is_explicit: bool = source.get("is_explicit", False)


SLOTS = (
    "title", "size", "ext", "url", "date", "file_type", "preview",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Doc(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/doc
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("doc", source)

        #: Title
        self.title: str = source.get("title", "")
        #: Size in bytes
        self.size: int = int(source.get("size", 0))
        #: File extension
        self.ext: str = source.get("ext", "")
        #: Url for downloading
        self.url: str = source.get("url", "")
        #: Upload date in unixtime
        self.date: int = int(source.get("date", 0))
        #: File type
        self.file_type: int = int(source.get("type", 0))
        #: Object with photo, graffiti or audio_message data for displaying
        #: preview for document. It's raw dictionary.
        self.preview: Dict = source.get("preview", {})


SLOTS = (
    "url", "title", "caption", "description", "photo",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Link(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/link
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("link", source)

        #: Url
        self.url: str = source.get("url", "")
        #: Title
        self.title: str = source.get("title", "")
        #: Caption if present
        self.caption: str = source.get("caption", "")
        #: Description
        self.description: str = source.get("description", "")
        #: Preview' if present
        self.photo: Optional[Photo] = None
        if "photo" in source:
            self.photo = Photo(source["photo"])


SLOTS = (
    "product_id", "sticker_id", "images", "images_with_background",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Sticker(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/sticker
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("sticker", source)

        #: Sticker pack's id
        self.product_id: int = int(source.get("product_id", 0))
        #: ID
        self.sticker_id: int = int(source.get("sticker_id", 0))
        #: Tuple with images without background
        self.images: Tuple[StickerSize, ...] = tuple(
            StickerSize(s) for s in source.get("images", ())
        )
        #: Tuple with images with background
        self.images_with_background: Tuple[StickerSize, ...] = tuple(
            StickerSize(s) for s in source.get("images_with_background", ())
        )


@read_only_properties("url", "width", "height")  # pylint: disable=too-few-public-methods
class StickerSize:
    """Class for storing information about sticker image."""

    def __init__(self, source):
        # Url
        self.url: str = source.get("url", "")
        # Width
        self.width: int = int(source.get("width", 0))
        # Height
        self.height: int = int(source.get("height", 0))


SLOTS = (
    "thumb_256", "thumb_96", "thumb_48"
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Gift(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/gift
    """

    __slots__ = tuple(SLOTS)

    def __init__(self, source: Dict):
        super().__init__("gift", source)

        #: Url to image with size 256x256
        self.thumb_256: str = source.get("thumb_256", "")
        #: Url to image with size 96x96
        self.thumb_96: str = source.get("thumb_96", "")
        #: Url to image with size 48x48
        self.thumb_48: str = source.get("thumb_48", "")


SLOTS = (
    "from_id", "date", "text", "reply_owner_id", "reply_post_id",
    "comments_count", "likes_count", "reposts_count", "views_count",
    "post_type", "post_source", "attachments", "copy_history",
)

@read_only_properties(*SLOTS)  # pylint: disable=too-few-public-methods,too-many-instance-attributes
class Wall(AttachmentContent):
    """
    Class for storing information about photo in attachment.
    Documentation: https://vk.com/dev/objects/post
    """

    __slots__ = SLOTS

    def __init__(self, source: Dict):
        super().__init__("wall", source)

        #: Post author
        self.from_id: int = int(source.get("from_id", 0))
        #: Publication date in unixtime
        self.date: int = int(source.get("date", 0))
        #: Text content
        self.text: str = source.get("text", "")
        #: In post is replying to post, it's owner id
        self.reply_owner_id: int = int(source.get("reply_owner_id", 0))
        #: In post is replying to post, it's post id
        self.reply_post_id: int = int(source.get("reply_post_id", 0))
        #: Comments count
        self.comments_count: int = int(source.get("comments", {}).get("count", 0))
        #: Likes count
        self.likes_count: int = int(source.get("likes", {}).get("count", 0))
        #: Reposts count
        self.reposts_count: int = int(source.get("reposts", {}).get("count", 0))
        #: Views count
        self.views_count: int = int(source.get("views", {}).get("count", 0))
        #: Post type
        self.post_type: str = source.get("post", "")
        #: Raw object with post source
        self.post_source: Dict = source.get("post_source", {})

        #: List of attachments
        self.attachments: Tuple[Attachment, ...] = tuple(
            Attachment(s) for s in source.get("attachments", ())
        )

        #: History of reposts
        self.copy_history: Tuple["Wall", ...] = tuple(
            Wall(s) for s in source.get("copy_history", ())
        )


@read_only_properties("type", "member_id", "text", "email", "photo")  # pylint: disable=too-few-public-methods
class Action:
    """Class for storing information about action in "message_new" event."""

    __slots__ = ("type", "member_id", "text", "email", "photo")

    def __init__(self, source: Dict):
        #: Action type
        self.type: str = source.get("type", "")
        #: Member id if action related to users
        self.member_id: int = int(source.get("member_id", 0))
        #: New title
        self.text: str = source.get("text", "")
        #: Invited person's email (if memeber_id < 0)
        self.email: str = source.get("email", "")
        #: Three sizes of new cover
        self.photo: Dict[str, str] = source.get("photo", {})



#!/usr/bin/env python3

from setuptools import setup


with open('README.md', "r") as f:
    long_description = f.read()


setup(
    name='vkpore',
    version='0.1.0',
    description='Library for organizing interactions with Vkontakte api.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=["library", "vkontakte", "asynchronous", "asyncio", "longpoll"],
    url='https://github.com/ekonda/vkpore/',
    author='Michael Krukov',
    author_email='krukov.michael@ya.ru',
    packages=['vkpore'],
    install_requires=[
        'aiohttp',
    ],
    python_requires='>=3.5',
    classifiers=[
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
    ]
)
